/* This file is an amalgamation of several source files.
   Please look for BEGIN FILE,END FILE and CONTINUE FILE
   strings to browse contents of individual files contained
   here. */
/* BEGIN FILE pbuffer.c */
#include <stdlib.h>
#include <string.h>
/* BEGIN FILE pbuffer.h */
typedef unsigned int color_t;
typedef struct
{
  int x,y;
// ARGB
  color_t color;
} point_t;
#define D_RED(x) ((x)>>16)&0xff)
#define D_GRE(x) ((x)>>8)&0xff)
#define D_BLU(x) ((x)>>0)&0xff)
#define D_ALP(x) ((x)>>24)&0xff)
typedef struct pbuffer pbuffer_t;
pbuffer_t *pbuffer_new();
void       pbuffer_free(pbuffer_t *);
void       pbuffer_clip(pbuffer_t *B,int x, int y, int w, int h);
pbuffer_t *pbuffer_get(pbuffer_t *B, point_t **Rp, unsigned int *Rl);
void       pixel(pbuffer_t* B, int x,int y,color_t color);
void       pbuffer_unify(pbuffer_t *B);
void       pbuffer_translate(pbuffer_t *B,int dx,int dy);
/* END FILE pbuffer.h */
/* CONTINUE FILE pbuffer.c */
#define NPO 50
typedef struct pnode
{
  struct pnode *next;
  point_t points[NPO];
  int     npoints;
} pnode_t;
struct pbuffer
{
  pnode_t *first, *last;
};
static pnode_t *pnode_new()
{
  return calloc(1,sizeof(pnode_t));
}
pbuffer_t *pbuffer_new()
{
  pbuffer_t *B;
  B= calloc(1,sizeof(*B));
  B->first= B->last= pnode_new();
  return B;
}
void pbuffer_free(pbuffer_t *B)
{
  pnode_t *P,*N;
  if (!B) return ;
  for(P=B->first;P;P=N)
  {
    N= P->next;
    free(P);
  }
  free(B);
}
void       pixel(pbuffer_t* B, int x,int y,color_t color)
{
  point_t po;
  po.x= x;
  po.y= y;
  po.color= color;
  if (B->last->npoints==NPO)
  {
    B->last->next= pnode_new();
    B->last= B->last->next;
  }
  B->last->points[B->last->npoints]= po;
  B->last->npoints++;
}
pbuffer_t* pbuffer_get(pbuffer_t *B, point_t **Rp, unsigned int *Rl)
{
  unsigned int np;
  pnode_t *P;
  point_t *D;
  np= 0;
  if (!B) { if (Rp) *Rp= NULL; if (Rl) *Rl= 0; return B; }
  for(P=B->first;P;P=P->next)
   np+= P->npoints;
  if (Rl) *Rl= np;
  if (Rp)
  {
    D= malloc(sizeof(point_t)*np);
    np= 0;
    for(P=B->first;P;P=P->next)
    {
      memcpy(D+np, P->points, sizeof(point_t)*P->npoints);
      np+= P->npoints;
    }
    *Rp= D;
  }
  return B;
}
void       pbuffer_clip(pbuffer_t *B,int x, int y, int w, int h)
{
  pnode_t *S;
  int i;
  if (!B) return ;
  for(S=B->first;S;S=S->next)
    for(i=0;i<S->npoints;i++)
      if (S->points[i].x<x ||
          S->points[i].y<y ||
          S->points[i].x>=(x+w) ||
          S->points[i].y>=(y+h))
      {
        S->points[i]= S->points[S->npoints-1];
        S->npoints--;
        i--;
      }
}
static int compoint(const void *pA, const void *pB)
{
  const point_t *A, *B;
  int R;
  A= pA; B= pB;
  R= A->x - B->x;
  if (R) return R;
  return A->y - B->y;
}
static void unify_block(pnode_t *P)
{
  int i,j;
  if (!P->npoints) return ;
  j= 0;
  for(i=1;i<P->npoints;i++)
    if (compoint(P->points+i, P->points+j))
    {
       j= j+1;
       if (i!=j) P->points[j]= P->points[i];
    }
  P->npoints= j+1;
}
static int find_point(point_t *P, int start, pnode_t *tab)
{
  int i,j,k,R;
  i= start;
  j= tab->npoints-1;
  while(i<=j)
  {
    k= (i+j)/2;
    R= compoint(P, tab->points+k);
    if (R>0) i= k+1;
    else if (R<0) j= k-1;
    else return k;
  }
  return -1;
}
static void unify_against(pnode_t *lop, pnode_t *fin)
{
  int nL, nF;
  int i,j,K,M;
  nL= lop->npoints;
  nF= fin->npoints;
  if (nL>nF) { unify_against(fin, lop); return ; }
  if (!nL || !nF) return ;
// completely disjoint
  if (lop->points[nL-1].x<fin->points[0].x) return ;
  if (lop->points[0].x>fin->points[nF-1].x) return ;
  j= -1;
  K= 0;
  for(i=0;i<nL;i++)
  {
    M= find_point(lop->points+i, K, fin);
    if (M==-1)
    {
      j++;
      if (i!=j) lop->points[j]= lop->points[i];
    }
    else K= M;
  }
  lop->npoints= j+1;
}
void       pbuffer_unify(pbuffer_t *B)
{
  pnode_t *P, *R;
  if (!B || !B->first) return ;
  for(P=B->first;P;P=P->next)
  {
    qsort(P->points, P->npoints, sizeof(P->points[0]), compoint);
    unify_block(P);
  }
  for(P=B->first;P!=B->last;P=P->next)
   for(R=P->next;R;R=R->next)
     unify_against(P,R);
}
void       pbuffer_translate(pbuffer_t *B,int dx,int dy)
{
  int i;
  pnode_t *P;
  if (!B) return;
  for(P=B->first;P;P=P->next)
    for(i=0;i<P->npoints;i++)
    {
     P->points[i].x += dx;
     P->points[i].y += dy;
    }
}
/* END FILE pbuffer.c */
/* BEGIN FILE line.c */
/* CONTINUE FILE line.c */
/* BEGIN FILE line.h */
pbuffer_t* line(pbuffer_t *B,unsigned int color,int x0, int y0, int x1, int y1);
/* END FILE line.h */
/* CONTINUE FILE line.c */
pbuffer_t* line(pbuffer_t *B,unsigned int color,int x0, int y0, int x1, int y1)
{
  int dx = abs(x1-x0), sx = x0<x1 ? 1 : -1;
  int dy = abs(y1-y0), sy = y0<y1 ? 1 : -1;
  int err = (dx>dy ? dx : -dy)/2, e2;
  B= B ? B : pbuffer_new();
  for(;;){
    pixel(B,x0,y0,color);
    if (x0==x1 && y0==y1) break;
    e2 = err;
    if (e2 >-dx) { err -= dy; x0 += sx; }
    if (e2 < dy) { err += dx; y0 += sy; }
  }
  return B;
}
/* END FILE line.c */
/* BEGIN FILE thick.c */
#include <math.h>
/* CONTINUE FILE thick.c */
/***********************************************************************
 *                                                                     *
 *                            X BASED LINES                            *
 *                                                                     *
 ***********************************************************************/
static void x_perpendicular(pbuffer_t *B, unsigned int color,
                            int x0,int y0,int dx,int dy,int xstep, int ystep,
                            int einit,int w_left, int w_right,int winit)
{
  int x,y,threshold,E_diag,E_square;
  int tk;
  int error;
  int p,q;
  threshold = dx - 2*dy;
  E_diag= -2*dx;
  E_square= 2*dy;
  p=q=0;
  y= y0;
  x= x0;
  error= einit;
  tk= dx+dy-winit;
  while(tk<=w_left)
  {
     pixel(B,x,y, color);
     if (error>=threshold)
     {
       x= x + xstep;
       error = error + E_diag;
       tk= tk + 2*dy;
     }
     error = error + E_square;
     y= y + ystep;
     tk= tk + 2*dx;
     q++;
  }
  y= y0;
  x= x0;
  error= -einit;
  tk= dx+dy+winit;
  while(tk<=w_right)
  {
     if (p)
       pixel(B,x,y, color);
     if (error>threshold)
     {
       x= x - xstep;
       error = error + E_diag;
       tk= tk + 2*dy;
     }
     error = error + E_square;
     y= y - ystep;
     tk= tk + 2*dx;
     p++;
  }
// we need this for very thin lines
  if (q==0 && p<2) pixel(B,x0,y0,color);
}
static void x_varthick_line
   (pbuffer_t *B, unsigned int color,
    int x0,int y0,int dx,int dy,int xstep, int ystep,
    double (*left)(void *,int ,int), void *argL,
    double (*right)(void *,int ,int),void *argR, int pxstep,int pystep)
{
  int p_error, error, x,y, threshold, E_diag, E_square, length, p;
  int w_left, w_right;
  double D;
  p_error= 0;
  error= 0;
  y= y0;
  x= x0;
  threshold = dx - 2*dy;
  E_diag= -2*dx;
  E_square= 2*dy;
  length = dx+1;
  D= sqrt(dx*dx+dy*dy);
  for(p=0;p<length;p++)
  {
    w_left=  (*left)(argL, p, length)*2*D;
    w_right= (*right)(argR,p, length)*2*D;
    x_perpendicular(B,color,x,y, dx, dy, pxstep, pystep,
                                      p_error,w_left,w_right,error);
    if (error>=threshold)
    {
      y= y + ystep;
      error = error + E_diag;
      if (p_error>=threshold)
      {
        x_perpendicular(B,color,x,y, dx, dy, pxstep, pystep,
                                    (p_error+E_diag+E_square),
                                     w_left,w_right,error);
        p_error= p_error + E_diag;
      }
      p_error= p_error + E_square;
    }
    error = error + E_square;
    x= x + xstep;
  }
}
/***********************************************************************
 *                                                                     *
 *                            Y BASED LINES                            *
 *                                                                     *
 ***********************************************************************/
static void y_perpendicular(pbuffer_t *B, unsigned int color,
                            int x0,int y0,int dx,int dy,int xstep, int ystep,
                            int einit,int w_left, int w_right,int winit)
{
  int x,y,threshold,E_diag,E_square;
  int tk;
  int error;
  int p,q;
  p=q= 0;
  threshold = dy - 2*dx;
  E_diag= -2*dy;
  E_square= 2*dx;
  y= y0;
  x= x0;
  error= -einit;
  tk= dx+dy+winit;
  while(tk<=w_left)
  {
     pixel(B,x,y, color);
     if (error>threshold)
     {
       y= y + ystep;
       error = error + E_diag;
       tk= tk + 2*dx;
     }
     error = error + E_square;
     x= x + xstep;
     tk= tk + 2*dy;
     q++;
  }
  y= y0;
  x= x0;
  error= einit;
  tk= dx+dy-winit;
  while(tk<=w_right)
  {
     if (p)
       pixel(B,x,y, color);
     if (error>=threshold)
     {
       y= y - ystep;
       error = error + E_diag;
       tk= tk + 2*dx;
     }
     error = error + E_square;
     x= x - xstep;
     tk= tk + 2*dy;
     p++;
  }
// we need this for very thin lines
  if (q==0 && p<2) pixel(B,x0,y0,color);
}
static void y_varthick_line
   (pbuffer_t *B, unsigned int color,
    int x0,int y0,int dx,int dy,int xstep, int ystep,
    double (*left)(void *,int ,int), void *argL,
    double (*right)(void *,int ,int),void *argR,int pxstep,int pystep)
{
  int p_error, error, x,y, threshold, E_diag, E_square, length, p;
  int w_left, w_right;
  double D;
  p_error= 0;
  error= 0;
  y= y0;
  x= x0;
  threshold = dy - 2*dx;
  E_diag= -2*dy;
  E_square= 2*dx;
  length = dy+1;
  D= sqrt(dx*dx+dy*dy);
  for(p=0;p<length;p++)
  {
    w_left=  (*left)(argL, p, length)*2*D;
    w_right= (*right)(argR,p, length)*2*D;
    y_perpendicular(B,color,x,y, dx, dy, pxstep, pystep,
                                      p_error,w_left,w_right,error);
    if (error>=threshold)
    {
      x= x + xstep;
      error = error + E_diag;
      if (p_error>=threshold)
      {
        y_perpendicular(B,color,x,y, dx, dy, pxstep, pystep,
                                     p_error+E_diag+E_square,
                                     w_left,w_right,error);
        p_error= p_error + E_diag;
      }
      p_error= p_error + E_square;
    }
    error = error + E_square;
    y= y + ystep;
  }
}
/***********************************************************************
 *                                                                     *
 *                                ENTRY                                *
 *                                                                     *
 ***********************************************************************/
pbuffer_t* draw_varthick_line
      (pbuffer_t *B, unsigned int color,
       int x0,int y0,int x1, int y1,
       double (*left)(void *,int ,int), void *argL,
       double (*right)(void *,int ,int), void *argR)
{
  int dx,dy,xstep,ystep;
  int pxstep, pystep;
// whether left and right get switched.
  int xch;
  dx= x1-x0;
  dy= y1-y0;
  xstep= ystep= 1;
  if (dx<0) { dx= -dx; xstep= -1; }
  if (dy<0) { dy= -dy; ystep= -1; }
  if (dx==0) xstep= 0;
  if (dy==0) ystep= 0;
  if (!B) B= pbuffer_new();
  if (dx==0 && dy==0) return B;
  xch= 0;
  switch(xstep + ystep*4)
  {
// -5
    case -1 + -1*4 :  pystep= -1; pxstep= 1; xch= 1; break;
// -1
    case -1 +  0*4 :  pystep= -1; pxstep= 0; xch= 1; break;
// 3
    case -1 +  1*4 :  pystep=  1; pxstep= 1; break;
// -4
    case  0 + -1*4 :  pystep=  0; pxstep= -1; break;
// 0
    case  0 +  0*4 :  pystep=  0; pxstep= 0; break;
// 4
    case  0 +  1*4 :  pystep=  0; pxstep= 1; break;
// -3
    case  1 + -1*4 :  pystep= -1; pxstep= -1; break;
// 1
    case  1 +  0*4 :  pystep= -1; pxstep= 0;  break;
// 5
    case  1 +  1*4 :  pystep=  1; pxstep= -1; xch=1; break;
  }
  if (xch) { void *K;
    K= argL; argL= argR; argR= K;
    K= left; left= right; right= K; }
  if (dx>dy) x_varthick_line(B,color,x0,y0,dx,dy,xstep,ystep,
                                                left,argL,right,argR,
                                                pxstep,pystep);
        else y_varthick_line(B,color,x0,y0,dx,dy,xstep,ystep,
                                                left,argL,right,argR,
                                                pxstep,pystep);
  return B;
}
static double cstval(void *arg, int dum1, int dum2)
{
  return *(double*) arg;
}
pbuffer_t* draw_thick_line
      (pbuffer_t *B, unsigned int color,
       int x0,int y0,int x1, int y1,
       double width)
{
  double L,R;
  L= width /2 ;
  R= width - L;
  return draw_varthick_line(B, color, x0, y0, x1, y1, cstval, &L, cstval, &R);
}
/* END FILE thick.c */
/* BEGIN FILE filledcirc.c */
pbuffer_t* draw_filled_circle
  (pbuffer_t *B, unsigned int color, int x, int y, int r)
{
    int i, ii, jj;
    if (!B) B= pbuffer_new();
    if (r < 0) r *= -1;
    ii = r; jj = 0;
    while (ii >= jj) {
 for (i = x - ii; i <= x + ii; i++) {
     pixel(B, i, y - jj,color);
     pixel(B, i, y + jj,color);
 }
 for (i = x - jj; i <= x + jj; i++) {
     pixel(B, i, y - ii, color);
     pixel(B, i, y + ii, color);
 }
 if ((r -= (jj++ << 1) - 1) < 0)
     r += (ii-- - 1) << 1;
    }
    return B;
}
/* END FILE filledcirc.c */
