typedef struct
      {
  int   type;
  int   x, y;
  int   modifiers;
  int   button;
  int   key;
  char *string;
  void *usr;
  int wide;
  double flo0, flo1;
  unsigned long time;
      }
    gfx_event_t;

enum
{
  GFXEV_BUTTON_PRESS=1, GFXEV_BUTTON_RELEASE, GFXEV_DOUBLE_CLICK,
  GFXEV_KEY_PRESS,      GFXEV_KEY_RELEASE,
  GFXEV_POINTER_MOVE,   
  GFXEV_READY_READ, GFXEV_READY_WRITE, GFXEV_READY_EXCEPT,
  GFXEV_QUIT, GFXEV_LAST,
};

#define GFXEV(X,Y) (((X)<<8)|(GFXEV_##Y))

#define GFXRESERVED 24

/* The event routine returns an event from the queue. It waits a maximum
   of 'msecs' milliseconds for the event. If msecs is zero, it waits
   until an event is received.

   The return value is one if an event is returned. It's zero if the
   return was due to timeout.

   In case of an event, *msecs is updated to reflect the remaining
   wait time.  */

int gfx_event(unsigned int *msecs,gfx_event_t *Revent);

enum
{ 
   GFX_K_PRT=1, 
   GFX_K_BACKSPACE,     GFX_K_TAB,      GFX_K_RETURN, GFX_K_ESCAPE,
   GFX_K_INSERT,        GFX_K_DELETE,   GFX_K_HOME,   GFX_K_END,
   GFX_K_PAGEUP,        GFX_K_PAGEDOWN,
   GFX_K_LEFT,          GFX_K_RIGHT,    GFX_K_DOWN,   GFX_K_UP,
   GFX_K_CONTROL=0x100, GFX_K_SHIFT= 0x200 , GFX_K_ALT= 0x400,
};

int gfx_init(char *name, int width, int height, char *title);
void gfx_close();


void gfx_geometry(unsigned int *W, unsigned int *H);
void gfx_monitor_file(int fd, int ident, int R,int W,int X);
void gfx_stop_monitor_file(int fd,int ident);
void gfx_leave_screen();
void gfx_set_state(int wno, int state);
int  gfx_get_state(int wno);
int  gfx_get_selected(int wno);
void gfx_set_selected(int wno,int sel);
void gfx_set_text(int wno,char *str);
void gfx_set_fgbg(int wno, int fg, int bg);
void gfx_insert(int wno,int pos, char *text,void *user,unsigned int* icon);
void gfx_remove(int wno, int pos, void **Ruser);
char* gfx_get_text(int wno);
void gfx_clear(int wno);
void gfx_focus(int wno);
void gfx_text_data(int font, const char *str, 
                   int *Rw, int *Rh,int *Ra,unsigned char **Rd);
void gfx_scrollbar_set(int wno, double start, double end, double step);
void gfx_pixgrid_attribs(int pg, int *Rshown, unsigned int *Rw, unsigned int *Rh);
void gfx_pixgrid_put(int pg, int x,int y,
                     unsigned int *img, unsigned int iW,unsigned int iH,
                     int src_x, int src_y, 
                     int copy_w, int copy_h);
void gfx_pixgrid_putsimple(int pg, unsigned int *img);
 
int gfx_ask(char *title, char *butstr, char *fmt, ...);
char* gfx_select_file(int mode,char *init,char **types);
void gfx_canvas_setsize(int wno, int new_w, int new_h);
void gfx_canvas_draw_rectangle(int wno,int color, int x,int y,int w,int h);
void gfx_canvas_fill_rectangle(int wno,int color, int x,int y,int w,int h);
void gfx_canvas_recsel_begin(int wno, gfx_event_t *ev);
void gfx_canvas_recsel_end(int wno,int *Rx1, int *Ry1, int *Rx2, int *Ry2);
void gfx_canvas_print(int font, int color, int rot, int x, int y, char *str);
void gfx_canvas_put_graymap
       (int wno, unsigned int color, int dx,int dy, int w, int h,
        unsigned char *source, int sw, int sh, int sx, int sy);
void gfx_canvas_put
       (int wno, int dx,int dy, int w, int h,
        unsigned int *source, int sw, int sh, int sx, int sy);
void gfx_spinner_set_imin(int wno, int min);
void gfx_spinner_set_imax(int wno, int max);
void gfx_spinner_set_iquant(int wno, int qua);
void gfx_spinner_set_fmin(int wno, double min);
void gfx_spinner_set_fmax(int wno, double max);
void gfx_spinner_set_fquant(int wno, double qua);
void gfx_spinner_set_fvalue(int wno, double value);
void gfx_spinner_set_ivalue(int wno, int value);
int gfx_spinner_get_ivalue(int wno);
double gfx_spinner_get_fvalue(int wno);
void gfx_logview_printf(int wno, const char *fmt, ...);
void gfx_logview_write(int wno, const char *txt, size_t len);
void gfx_logview_clear(int wno);
void gfx_slider_set(int wno, double pos);
double gfx_slider_get(int wno);
void gfx_filelist_set_directory(int wno,char *dir);
void gfx_filelist_set_types(int wno, char **types);
char** gfx_filelist_get_files(int wno);
char *gfx_filelist_get_directory(int wno);

#ifdef GFX_ANDROID
JavaVM *gfx_get_javavm();
#endif

enum { _XX_WIDGETS_START=GFXEV_LAST+GFXRESERVED,
  /* WIDGET NAMES BELOW */
  Wuseless,
  Wcanvas,
  Wremove,
  Wrandomprefix,
  Wstats,
  Wnormalize,
  Wsave,
  Wdensify,
  Wsorts,
  Wremovefinal,
  Wrnormalize,
  Wtryremove,
  Wreproduce,
  Wflipuseless,
  Wswap,
  Wrmsingle,
  Wfindshuf,
  Wshuf,
  Wrepromax,
  Wwall,
  Wtryfwd,
  Wprepwall,
  Wdouble,
  Whalf,
  Wninputs,
  Wstartgen,
  Wmkgate,
  Wcrazymax,
  Wcrazymin,
  Wscanposfwd,
  Wgate,
  Wmkthis,

 _XX_WIDGETS_END };
void gfx_SCREEN_0();
