#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include "gfx.h"
#include "network.h"
#include "draw-net.h"
#include "main.h"
#include "manser.h"
#include "intset.h"
#include "util.h"

network_t *NW;

typedef struct { int a,b; int metric; } pair_t;


static void show_stats()
{
  char buf[200];
  if (!NW->ngates)
    sprintf(buf,"empty network");
  else
    sprintf(buf,"%d gates %d levels",
        NW->ngates, NW->gates[NW->ngates-1]->level+1);
  gfx_set_text(Wstats,buf);
}

void show_network(int siz)
{
  int w,h;
  unsigned int *data;
  if (!NW) return ;
  draw_network(NW, 30,30,7,3,&w,&h,&data);
  if (siz)
    gfx_canvas_setsize(Wcanvas, w,h);
  gfx_canvas_put(Wcanvas,0,0,w,h,data,w,h,0,0);
  free(data);
  show_stats();
}

void replace_network(network_t *WW)
{
  if (NW) network_free(NW);
  NW= WW;
  show_network(1);
}

static int number_of_combinations(int N,int M)
{
  return factorial(N) / factorial(M) / factorial(N-M);
}

static int NI;
static exp_t **final;
static int **rela;
static int *comp;

static exp_t *make_final_exp(int N, int z)
{
  int M;
  exp_t *E;
  int value;
  int i[64];
  int depth;

  M= N-z;
  E= exp_new(number_of_combinations(N, M));

  depth= 0;
  value= 0;
  i[0]= N-1;

restart:
  if (M==depth)
  {
    E->data[E->len++]= value;
    goto ret;
  }
loop_next:
  if (i[depth]-(M-depth)<-1) goto ret;
  value= value | (1<<i[depth]);
  depth++;
  i[depth]= i[depth-1]-1;
  goto restart;
loop_cont:
  value= value & ~(1<<i[depth]);
  i[depth]--;
  goto loop_next;
ret:
  if (depth!=0)
  {
    depth--;
    goto loop_cont;
  }

  E->data[E->len]= 0;
  return E;
}

static void make_final(int n)
{
  int i;

  if (rela)
  {
    for(i=0;i<NI+2;i++) free(rela[i]);
    free(rela);
    rela= NULL;
  }

  if (final)
  {
    for(i=0;i<NI;i++) exp_free(final[i]);
    free(final);
    final= NULL;
  }
  NI= n;

  final= malloc(sizeof(*final)*NI);
  for(i=0;i<NI;i++)
    final[i]= make_final_exp(NI, i);

  if (comp) free(comp);
  comp= calloc(sizeof(comp[0]),NI);

  rela= malloc(sizeof(rela[0])*(NI+2));
  for(i=0;i<NI+2;i++) rela[i]= calloc(sizeof(rela[i][0]),NI+2);
}

static void print_rela(int ex)
{
  int i; char *z;
  int n;
  z= "<.>";
  int j;
  printf("Ordering Relation%s\n\n",ex? " (with extras)":"");
  if (ex) n= NI+2; else n= NI;
  printf("  "); for(i=0;i<n;i++) printf("%d",i%10); printf("\n");
  for(i=0;i<n;i++)
  {
    printf(" %d",i%10);
    for(j=0;j<n;j++)
      printf("%c",z[1+rela[i][j]]);
    printf("\n");
  }
  
}

static void print_final()
{
  int i;  char *z;
  if (!final) return ;
  printf("final expressions\n");
  printf("-----------------\n");
  for(i=0;i<NI;i++)
  {
    printf("%2d: ", i);
    z= exp_to_string(final[i]);
    puts(z);
    free(z);
  }

  print_rela(0);
}

static uint32_t maximal(uint32_t a)
{
  int i,j;
  for(i=0;i<NI;i++)
    if (a&(1<<i))
     for(j=0;j<NI;j++)
      if (rela[j][i]==1) a |= (1<<j);
  return a;
}

static uint32_t minimal(uint32_t a)
{
  int i,j;
  for(i=0;i<NI;i++)
    if (a&(1<<i))
     for(j=0;j<NI;j++)
      if (rela[j][i]==1) a &= ~(1<<j);
  return a;
}

char *term2str(uint32_t K)
{
  int i;
  static char b[200];
  char *m="0123456789abcdefghijklmnopqrstuvwxyz";
  int c;
  for(i=0,c=0;i<32;i++) if (K&(1<<i)) b[c++]= m[i];
  b[c]= 0;
  return b;
}

static void  combine_terms_normal(exp_t *dst, exp_t *src,uint32_t a,uint32_t b)
{
  int i;
  intset_t *A,*B, *find, *store;
  uint32_t V,R,ZZ;
  int ca,cb;
  char *z;

  A= intset_new();
  B= intset_new();

  for(i=0;i<src->len;i++)
  {
    V= src->data[i];
    R= V & (a|b);
    if (R==0 || R==(a|b)) { dst->data[dst->len++]= V & ~b;  continue; }
    else if (R==a)        { find= B; store= A; }
    else                  { find= A; store= B; }
    printf("  term: %s ", term2str(V));
    V &= ~(a|b);
    
    if (intset_find(find, V))
    {
       printf("not found, storing\n");
       intset_insert(store, V);
    }
    else
    {
       printf("found, combining\n");
       intset_remove(find, V);
       dst->data[dst->len++]= V | b;
    }
  }

  src->len= 0;
  for(ca=0;!intset_destroy(A,&ZZ);ca++) src->data[src->len++]= ZZ | a; 
  for(cb=0;!intset_destroy(B,&ZZ);cb++) src->data[src->len++]= ZZ | b;

  free(A); free(B);
  printf(" combined terms normally: %d left in A, %d left in B\n", ca,cb);
  if (ca || cb)
  {
    z= exp_to_string(src);
    printf(" remaining expression: %s\n", z);
    free(z);
  }
}

static int rewrite_term_rule4(uint32_t *R,uint32_t V, int a, int b)
{
  int i;
  if ((V&(1<<b))==0) return 0;
  for(i=0;i<NI;i++) 
    if ( (rela[i][a]==-1) && (V&(1<<i)) )
    {
      *R= (V&~(1<<b)) | (1<<a);
      return 1;
    }
  return 0;
}

static int rewrite_term_rule5(uint32_t *R,uint32_t V, int a, int b)
{
  int i;
  if ((V&(1<<a))==0) return 0;
  for(i=0;i<NI;i++) 
    if ( (rela[i][b]==-1) && (V&(1<<i)) )
    {
      *R= V;
      return 1;
    }
  return 0;
}


static void rewrite_terms_rule45(exp_t *dst, exp_t *src, int a, int b)
{
  int i;
  uint32_t K;
  for(i=0;i<src->len;i++)
  {
    if ( rewrite_term_rule4(&K, src->data[i], a, b) ||
         rewrite_term_rule5(&K, src->data[i], a, b) )
         
    {
      src->data[i]= src->data[src->len-1];
      src->len--;
      i--;
      dst->data[dst->len++]= K;
    }
  }
}

static int greater_by_rule_8(uint32_t B, uint32_t A)
{
  int i,j;
  // the direction of the loop for Bi is important. it has
  // to go from lower values to higher values so that 
  // the relations between Bi also affect each other.
  for(i=0;i<NI;i++)
    if (B&(1<<i))
    {
      for(j=0;j<NI;j++)
        if (A&(1<<j))
        {
          if (j==i || rela[i][j]==1) { B&= ~(1<<i); A|= (1<<i); break; }
        }
    }
  if (B) return 0;
  return 1;
}

//
// if we have A + B where B > A, we convert it to simply B.
//
// determining B >A is difficult, but not impossible.
//
// For bi, if there exists an aj s.t. bi>aj, then
// we can append bi to A without changing its result
//
// if we can do this for all bi, then we have transformed
// the expression to A.B + B, which is just B.
//
static void simplify_by_rule8(exp_t *E)
{
  char *rem;
  int i,j;
  char *z;
  z= exp_to_string(E);
  printf("#####################\n");
  printf("Simplify rule8: %s\n",z); free(z);

  input_t gsto[2], fsto[2];
  exp_t G; G.len= 1; G.data= gsto; G.data[1]= 0;
  exp_t F; F.len= 1; F.data= fsto; F.data[1]= 0;
  char *z1,*z2;
  
  rem= calloc(sizeof(rem[0]),E->len);
  for(i=0;i<E->len;i++)
    for(j=0;j<E->len;j++)
     if (i!=j && greater_by_rule_8(E->data[j], E->data[i]))
     {
        rem[i]= 1;

        G.data[0]= E->data[i]; F.data[0]= E->data[j];
        z1= exp_to_string(&G); z2= exp_to_string(&F);
        printf("   Removing term %s due to term %s\n", z1,z2);
        free(z1); free(z2);
        break;
     }
  for(i=0,j=0;i<E->len;i++)
    if (!rem[i]) E->data[j++]= E->data[i];
  E->data[j]= 0;
  E->len= j;
  free(rem);
  z= exp_to_string(E);
  printf("result   rule8: %s\n",z); free(z);
  printf("#####################\n");
}

static void  combine_terms_max(exp_t *dst, exp_t *src,uint32_t a,uint32_t b)
{
  int i;
  intset_t *A,*B, *find, *store;
  uint32_t V,R,ZZ,Z;
  int ca,cb;
  char *z;

  A= intset_new();
  B= intset_new();

  for(i=0;i<src->len;i++)
  {
    V= src->data[i];
    R= V & (a|b);
    if (R==0 || R==(a|b)) { dst->data[dst->len++]= V & ~b; continue; }
    else if (R==a)        { find= B; store= A; }
    else                  { find= A; store= B; }
    Z= maximal(V);
    printf(" term: %s ", term2str(V));
    V &= ~(a|b);
    Z &= ~(a|b);
    printf(" max= %s ", term2str(Z));
    
    if (intset_find(find, Z))
    {
       printf("not found, storing\n");
       intset_insert(store, Z);
    }
    else
    {
       printf("found, combining\n");
       intset_remove(find, Z);
       dst->data[dst->len++]= V | b;
    }
  }

  src->len= 0;
  for(ca=0;!intset_destroy(A,&ZZ);ca++) src->data[src->len++]= minimal(ZZ | a); 
  for(cb=0;!intset_destroy(B,&ZZ);cb++) src->data[src->len++]= minimal(ZZ | b);

  free(A); free(B);
  printf(" combined terms maximized: %d left in A, %d left in B\n", ca,cb);
  if (ca || cb)
  {
    z= exp_to_string(src);
    printf("remaining expression: %s\n", z);
    free(z);
  }
}

static int   combine_terms_rule6
   (exp_t *dst, uint32_t Ca, uint32_t Db, int a,int b)
{
  uint32_t C,D,F,G,R;

  if (Ca&(1<<b) && (Db&(1<<a)))
  {
     C= Ca; Ca= Db; Db= C;
  }
  else if (!(Ca&(1<<a) && (Db&(1<<b)))) return 0;

  C= Ca & ~(1<<a);
  D= Db & ~(1<<b);

  F= maximal(Ca);
  G= maximal(Db);

  R= F & G;
  if ( ((R&C)==C) && ((R&D)==D))
  {
    dst->data[dst->len++]= minimal(R) | (1<<b);
    return 1;
  }
  return 0;
}

static void  rewrite_expr_rule6(exp_t *dst, exp_t *src,int a, int b)
{
  char *rem;
  int i,j;

  rem= calloc(src->len+1,sizeof(char));
  for(i=0;i<src->len;i++)
     for(j=0;j<src->len;j++)
     {
       if (i==j) continue;
       if (rem[i] && rem[j]) continue;
       if (combine_terms_rule6(dst, src->data[i], src->data[j], a,b))
       {
          rem[i]= rem[j]= 1;
       }
     }
  j= 0;
  for(i=0;i<src->len;i++)
   if (!rem[i]) src->data[j++]= src->data[i];
  src->data[j]= 0;
  src->len= j;
  free(rem);
}


static exp_t* rewrite_final_expr(exp_t *E,int ba,int bb)
{
  exp_t *cp,*res;
  uint32_t a,b;
  char *z;

  res= exp_new(E->len);

  cp= exp_new(E->len);
  cp->len= E->len;
  memcpy(cp->data, E->data, sizeof(E->data[0])*(E->len+1));
  E= cp;

  z= exp_to_string(E);
  printf("rewriting for gate(%d,%d): %s\n", ba,bb, z); free(z);

  a= (1<<ba);
  b= (1<<bb);

  combine_terms_normal(res, E, a,b);
  // combine_terms_max(res, E, a,b);

  rewrite_terms_rule45(res, E, ba, bb);
  rewrite_expr_rule6(res, E, ba, bb);

  if (E->len)
  {
    z= exp_to_string(E);
    printf("remaining for gate(%d,%d): %s\n", ba, bb, z); free(z);
    exp_free(res);
    res= NULL;
  }
  else
  {
    z= exp_to_string(res);
    printf("result    for gate(%d,%d): %s\n", ba, bb, z); free(z);
  }

  exp_free(E);
  return res;
}

static int compar_metric(const void *_a, const void *_b)
{
  const pair_t *A= _a;
  const pair_t *B= _b;
  return A->metric - B->metric;
}
static void sort_by_metric(pair_t *pairs, int npairs)
{
  qsort(pairs, npairs, sizeof(pairs[0]), compar_metric);
}

static void filter_by_metric(pair_t *pairs, int *metric, int *ioN)
{
  int N,i;
  int min;
  N= *ioN;

  min= metric[0];
  for(i=1;i<N;i++)
    min= imin2(metric[i], min);
  for(i=0;i<N;i++)
   if (metric[i]!=min)
   {
     metric[i]= metric[N-1];
     pairs[i]= pairs[N-1];
     i--;
     N--;
   }
  *ioN= N;
}

static void carry_rela_transitions()
{
  int i,j,k;
  for(k=0;k<NI;k++)
   for(i=0;i<NI;i++)
    for(j=0;j<NI;j++)
      if (rela[i][k]==-1 && rela[k][j]==-1) rela[i][j]= -1;
      else if (rela[i][k]==1 && rela[k][j]==1) rela[i][j]= 1;
}

static void print_network_end()
{
  printf("Current values at the end of the network\n");
  int i;
  for(i=0;i<NI;i++)
  {
    char *z;
    z= exp_to_string(NW->G[NW->ngates][i]);
    printf("%2d %s\n", i,z);
    free(z);
  }
}

static int try_gate(int a,int b)
{
  exp_t **newfin;
  int i,j,k;
  newfin= calloc(sizeof(*newfin), NI);

  for(i=0;i<NI;i++)
    if (!(newfin[i]=rewrite_final_expr(final[i], a,b)))
       break;

  if (i!=NI)
  {
    for(i=0;i<NI;i++) if (newfin[i]) exp_free(newfin[i]);
    free(newfin);
    return 0;
  }

  for(i=0;i<NI;i++) exp_free(final[i]);
  free(final);
  final= newfin;
  network_append_gate(NW, a,b);
  show_network(1);

  int ap,bp;
  ap= NI;
  bp= NI+1;

  for(i=0;i<NI+2;i++) { rela[ap][i]= 0; rela[i][ap]= 0; 
                        rela[bp][i]= 0; rela[i][bp]= 0; }

  printf("prior to adjustment\n");
  print_rela(0);

  for(k=0;k<NI;k++)
  {
    if (rela[a][k]==-1 && rela[b][k]==-1)
      { rela[ap][k]= rela[bp][k]= -1; rela[k][ap]= rela[k][bp]= 1; continue; }
    if (rela[a][k]==1 && rela[b][k]==1)
      { rela[ap][k]= rela[bp][k]= 1; rela[k][ap]= rela[k][bp]= -1; continue; }
    if (rela[a][k]== 1) { rela[bp][k]=  1; rela[k][bp]= -1; 
                          rela[ap][k]= 0; rela[k][ap]= 0; }
    if (rela[b][k]==-1) { rela[ap][k]= -1; rela[k][ap]=  1;
                          rela[bp][k]= 0; rela[k][bp]= 0; }
  }

  printf("after adjustment\n");
  print_rela(0);

  for(i=0;i<NI;i++) { rela[a][i]= rela[ap][i]; rela[i][a]= rela[i][ap];
                      rela[b][i]= rela[bp][i]; rela[i][b]= rela[i][bp]; }

  printf("after transfer\n"); print_rela(1);

  comp[b]= comp[a]= imax2(comp[a], comp[b]) + 1;

  rela[a][b]= -1;
  rela[b][a]=  1;

  printf("after addition\n"); print_rela(0);

  for(i=0;i<NI;i++)
   for(j=i+1;j<NI;j++)
   if (1 || !rela[i][j])
   {
      int R;
      R= exp_compare_value(NW->G[NW->ngates][i],NW->G[NW->ngates][j]);
      if (!R) continue;
      rela[i][j]=  R;
      rela[j][i]= -R;
   }

  carry_rela_transitions();
  printf("after transitions\n"); print_rela(0);

  printf("after adding relations derived from expressions\n"); print_rela(0);


  carry_rela_transitions();
  printf("after second wave of transitions\n"); print_rela(0);

  print_final();
  //print_network_end();
  return 1;
}

static void mkgate()
{
  int i,j,a,b;
  pair_t *pairs;
  int npairs;
  if (!NI || !NW) return ;

  pairs= malloc(sizeof(pairs[0])*NI*(NI-1)/2);
  npairs= 0;

  for(i=0;i<NI;i++)
   for(j=i+1;j<NI;j++)
   {
     if (rela[i][j] || rela[j][i]) continue;
     pairs[npairs].a= i;
     pairs[npairs].b= j;
     npairs++;
   }

  for(i=0;i<npairs;i++)
    pairs[i].metric= imax2(comp[pairs[i].a],comp[pairs[i].b])*1000 /
                            (pairs[i].b-pairs[i].a);

  sort_by_metric(pairs, npairs);

#if 0
  for(i=0;i<npairs;i++)
    metric[i]= pairs[i].b - pairs[i].a;

  filter_by_metric(pairs, metric, &npairs);
#endif

  for(i=0;i<NI;i++)
  {
   for(j=0;j<final[i]->len;j++)
     final[i]->data[j]= minimal(final[i]->data[j]);
   simplify_by_rule8(final[i]);
   exp_simplify(final[i]); // rule-1
  }

  printf("available pairs:");
  for(i=0;i<npairs;i++) printf(" %d.%d", pairs[i].a, pairs[i].b);
  printf("\n");

  for(i=0;i<npairs;i++)
  {
    a= pairs[i].a;
    b= pairs[i].b;
 
    if (try_gate(a,b)) 
    { 
      printf("success at %d.%d\n",a,b);
      break;
    }
  }
  free(pairs);
}


/***********************************************************************
 *                                                                     *
 *                              CRAZY ONE                              *
 *                                                                     *
 ***********************************************************************/

static int is_trivial(int num, exp_t *E)
{
  uint32_t V;
  if (E->len!=1) return 0;
  V= E->data[0];
  if (!(V&(1<<num))) return 0;
  V &= ~(1<<num);
  if (V) return 0;
  return 1;
}


/***********************************************************************
  returns 1 iff E is of the form:
     num: max ( num, num-1, num-2, .. num-l)
  num-l is returned in *Rl if successful.
 ***********************************************************************/
static int is_maximal_crazy_expr(int num, exp_t *E,int *Rl)
{
  int i,H,L;
  uint32_t V,vis;
  vis= 0;
  for(i=0;i<E->len;i++)
  {
   V= E->data[i];
   if (bit_count(V)!=1) return 0;
   vis |= V;
  }
  H= highest_bit(vis);
  L= lowest_bit(vis);
  if (H!=num || (H-L+1)!=E->len) return 0;
  *Rl= L;
  return 1;
}

int rule7(uint32_t A, uint32_t B)
{
  int i,j;
  for(i=0;i<NI;i++)
   if (A&(1<<i))
   {
     for(j=0;j<NI;j++)
       if (B&(1<<j)) { if (rela[i][j]!=-1) break; }
     if (j==NI) A &= ~(1<<i);
   }
  if (A) return 0;
  return 1;
}

void apply_rule_7(exp_t *E)
{
  int i,j;
  char *rem;
  rem= calloc(sizeof(rem[0]),E->len);
  for(i=0;i<E->len;i++)
   for(j=0;j<E->len;j++)
     if (i!=j && rule7(E->data[i],E->data[j]))
     {
       rem[i]= 1;
       break;
     }
  for(i=0,j=0;i<E->len;i++)
    if (!rem[i]) E->data[j++]= E->data[i];
  E->data[j]= 0;
  E->len= j;
}

void try_maximal_crazy()
{
  exp_t *E;
  int i,hi,lo,j;
  uint32_t mask;

  printf("looking for a maximal crazy expr\n");

  for(i=NI-1;i>0;i--)
   if (!is_trivial(i, final[i])) break;
  if (i==0) return ;

  if (!is_maximal_crazy_expr(i,final[i],&lo)) return ;
  hi= i;
  if (hi-lo!=1) return ;
  printf("found maximal crazy between %d-%d\n", lo,hi);

  mask= (1<<hi);

  printf("updating final expressions\n");
  for(i=0;i<=lo;i++)
  {
    E= final[i];
    for(j=0;j<E->len;j++)
      E->data[j] &= ~mask;
    exp_simplify(E);
  }

  final[hi]->data[0]= mask;
  final[hi]->data[1]= 0;
  final[hi]->len= 1;

  for(i=0;i<=lo;i++)
    { rela[i][hi]= -1; rela[hi][i]= 1; }

  printf("---done with the maximal crazy stuff. let's look at it\n");
  print_final();

#if 0
  printf("+++after applying rule 7:\n");
  for(i=0;i<NI;i++) apply_rule_7(final[i]);
  print_final();
#endif

  network_append_gate(NW, lo,hi);
  show_network(1);
}


/***********************************************************************
 *                                                                     *
 *                              CRAZY TWO                              *
 *                                                                     *
 ***********************************************************************/
static int is_minimal_crazy_expr(int num, exp_t *E,int *Rh)
{
  int H,L;
  uint32_t V;
  if (E->len!=1) return 0;
  V= E->data[0];
  H= highest_bit(V);
  L= lowest_bit(V);
  if (L!=num || (H-L+1)!=bit_count(V)) return 0;
  *Rh= H;
  return 1;
}

void try_minimal_crazy()
{
  int i,j,lo,hi;
  uint32_t mask;
  exp_t *E;

  printf("looking for a minimal crazy expression\n");
  for(i=0;i<NI-1;i++)
   if (!is_trivial(i,final[i])) break;
  if (i==NI-1) return ;

  lo= i;
  if (!is_minimal_crazy_expr(lo,final[lo],&hi)) return ;
  if (hi-lo!=1) return ;

  printf("found minimal crazy between %d-%d\n", lo,hi);

  mask= (1<<lo) | (1<<hi);

  printf("updating final expressions\n");
  for(i=hi;i<NI;i++)
  {
    E= final[i];
    for(j=0;j<E->len;j++)
    {
      if (E->data[j]&(1<<lo)) { E->data[j]= E->data[E->len-1];
                                j--; E->len--; }
    }
    exp_simplify(E);
  }

  final[lo]->data[0]= (1<<lo);
  final[lo]->data[1]= 0;
  final[lo]->len= 1;

  for(i=hi;i<NI;i++)
    { rela[i][lo]= 1; rela[lo][i]= -1; }

  printf("---done with the minimal crazy stuff. let's look at it\n");
  print_final();

  printf("+++after applying rule 7:\n");
  for(i=0;i<NI;i++) apply_rule_7(final[i]);
  print_final();

  network_append_gate(NW,lo,hi);
  show_network(1);

}

void scan_pos_fwd()
{
  uint32_t *masks;
  int *highest;
  int i,j;
  masks= calloc(sizeof(masks[0]),NI);
  highest= calloc(sizeof(highest[0]),NI);

  for(i=0;i<NI;i++)
   for(j=0;j<final[i]->len;j++)
    masks[i] |= final[i]->data[j];

  for(i=0;i<NI;i++)
  {
    for(j=NI-1;j>=0;j--)
      if (masks[j]&(1<<i)) break;
    highest[i]= j;
    if (j!=-1)
      printf("latest mention of %d is at final[%d]\n",i,j);
  }

  for(i=0;i<NI;i++)
   if (highest[i]!=-1)
     for(j=highest[i]+1;j<NI;j++)
       { rela[i][j]= -1; rela[j][i]= 1; }

  printf("results of forward position scan:\n");
  print_rela(0);
  free(highest); free(masks);
}

void make_this()
{
  int a,b;
  char *z,*e;
  z= gfx_get_text(Wgate);
  if (!z || !z[0]) return ;
  a= strtol(z, &e, 0);
  if (e==z) return ;
  z= e+1;
  b= strtol(z, &e, 0);
  if (e==z) return ;
  if (a<0 || b<0 || a>=NI || b>=NI) return ;

  try_gate(imin2(a,b), imax2(a,b));
}

int main(int argc,char **argv)
{
  char *k;
  int i;
  network_t *WW;

  if (argc>2) NW= network_from_cmdline(argv[1],argv[2]);
         else NW= network_from_file(argv[1]);
  gfx_init(NULL, 1000, 800, "Sorting Network Tool");
  gfx_SCREEN_0();
  show_network(1);
  while(1)
  {
    gfx_event_t event;
    if (!gfx_event(NULL,&event)) continue;
    if (bottom_event(event)) continue;
    switch(event.type)
    {
    case GFXEV_QUIT:
       goto done;
    case Wstartgen:
       k= gfx_get_text(Wninputs);
       if (!k) k= "2";
       i= atoi(k); if (i<2) i= 2;
       WW= network_new(i);
       replace_network(WW);

       make_final(i); print_final();
       break;
    case Wcrazymax: try_maximal_crazy(); break;
    case Wcrazymin: try_minimal_crazy(); break;
    case Wscanposfwd: scan_pos_fwd(); break;
    case Wmkgate: mkgate(); break;
    case Wmkthis: make_this(); break;
    }
  }
done:
  return 0;
}
