
typedef unsigned short input_t;

typedef struct 
{
  int A,B,x,level;
} gate_t;

typedef struct
{
  int ninputs,ngates;
  gate_t *gates;
} gatelist_t;

typedef struct
{
  int count;
  int len;
  input_t *data;
} exp_t;

typedef struct
{
  int A,B;
  exp_t *min,*max;

  int x; // for drawing..
  int color;

  int level;
} compex_t;

typedef struct
{
  int ninputs;
  int ngates,sgates;
  compex_t **gates;
     // G= ninputs * (ngates+1)
  exp_t ***G;
     //
     // first index of G is the gate number. gates are numbered
     // from 1 in this array. G[0] is the initial state where
     // G[0][x] = Ax  (input)
     // The second one is the line number, starting from 0.
     //
  int direction;
} network_t;


network_t *network_new(int ninputs);
void network_free(network_t *NW);
void network_append_gate(network_t *NW, int A,int B);
int network_sorts(network_t *NW);


network_t *network_from_cmdline(char *a1,char *a2);
network_t *network_from_binary(int len, unsigned char *data);
void network_to_binary(network_t *NW,int *Rlen, unsigned char **Rdata);
network_t *network_from_file(char *fn);
void network_to_file(network_t *NW,char *fn);

void network_layout_for_drawing(network_t *NW);
char *exp_to_string(exp_t *A);

int network_find_useless_gate(network_t *NW);
int network_find_useless_gate_reverse(network_t *NW);
int network_levels(network_t *NW);

void network_all_useless_gates(network_t *NW,int *Rn, int *Ru);

void network_normalize(network_t *NW);

void network_to_gate_list(network_t *NW,gatelist_t **rGL);
network_t* network_from_gate_list(gatelist_t *GL,int right);
void gatelist_free(gatelist_t *GL);

int network_find_shuffle_gate(network_t *NW);


exp_t *exp_min(exp_t *A,exp_t *B);
exp_t *exp_max(exp_t *A,exp_t *B);
// this is a lexical comparison, just like strcmp
int exp_compare(exp_t *A,exp_t *B);

// this is a value comparison. max(A,B) is computed
// to find out whether it's equal to A or B.
// 0 doesn't mean that the expressions are equal
// but that it's impossible to tell which is
// greater.
int exp_compare_value(exp_t *A,exp_t *B);

void exp_free(exp_t *A);


void exp_write(FILE* f, exp_t *E);
exp_t* exp_read(FILE* f);
exp_t* exp_new(int nterms);


int factorial(int N);

// this one uses rule-1
void exp_simplify(exp_t *E);


